# -*- coding: utf-8 -*-
from .utils import load
from .categories import unique_aliases, alias


confusables_data = load('confusables.json')


class Found(Exception):
    pass


def is_mixed_script(string, allowed_aliases=['COMMON']):
    """Checks if ``string`` contains mixed-scripts content, excluding script
    blocks aliases in ``allowed_aliases``.

    E.g. ``B. C`` is not considered mixed-scripts by default: it contains characters
    from **Latin** and **Common**, but **Common** is excluded by default.

    >>> confusables.is_mixed_script('Abç')
    False
    >>> confusables.is_mixed_script('ρτ.τ')
    False
    >>> confusables.is_mixed_script('ρτ.τ', allowed_aliases=[])
    True
    >>> confusables.is_mixed_script('Alloτ')
    True

    :param string: A unicode string
    :type string: str
    :param allowed_aliases: Script blocks aliases not to consider.
    :type allowed_aliases: list(str)
    :return: Whether ``string`` is considered mixed-scripts or not.
    :rtype: bool
    """
    allowed_aliases = [a.upper() for a in allowed_aliases]
    cats = unique_aliases(string) - set(allowed_aliases)
    return len(cats) > 1


def is_confusable(string, greedy=False, preferred_aliases=[]):
    """Checks if ``string`` contains characters which might be confusable with
    characters from ``preferred_aliases``.

    If ``greedy=False``, it will only return the first confusable character
    found without looking at the rest of the string, ``greedy=True`` returns
    all of them.

    ``preferred_aliases=[]`` can take an array of unicode block aliases to
    be considered as your 'base' unicode blocks:

    -  considering ``paρa``,

       -  with ``preferred_aliases=['latin']``, the 3rd character ``ρ``
          would be returned because this greek letter can be confused with
          latin ``p``.
       -  with ``preferred_aliases=['greek']``, the 1st character ``p``
          would be returned because this latin letter can be confused with
          greek ``ρ``.
       -  with ``preferred_aliases=[]`` and ``greedy=True``, you'll discover
          the 29 characters that can be confused with ``p``, the 23
          characters that look like ``a``, and the one that looks like ``ρ``
          (which is, of course, *p* aka *LATIN SMALL LETTER P*).

    >>> confusables.is_confusable('paρa', preferred_aliases=['latin'])[0]['character']
    'ρ'
    >>> confusables.is_confusable('paρa', preferred_aliases=['greek'])[0]['character']
    'p'
    >>> confusables.is_confusable('Abç', preferred_aliases=['latin'])
    False
    >>> confusables.is_confusable('AlloΓ', preferred_aliases=['latin'])
    False
    >>> confusables.is_confusable('ρττ', preferred_aliases=['greek'])
    False
    >>> confusables.is_confusable('ρτ.τ', preferred_aliases=['greek', 'common'])
    False
    >>> confusables.is_confusable('ρττp')
    [{'homoglyphs': [{'c': 'p', 'n': 'LATIN SMALL LETTER P'}], 'alias': 'GREEK', 'character': 'ρ'}]

    :param string: A unicode string
    :type string: str
    :param greedy: Don't stop on finding one confusable character - find all of them.
    :type greedy: bool
    :param preferred_aliases: Script blocks aliases which we don't want ``string``'s characters
        to be confused with.
    :type preferred_aliases: list(str)
    :return: False if not confusable, all confusable characters and with what they are confusable
        otherwise.
    :rtype: bool or list
    """
    preferred_aliases = [a.upper() for a in preferred_aliases]
    outputs = []
    checked = set()
    for char in string:
        if char in checked:
            continue
        checked.add(char)
        char_alias = alias(char)
        if char_alias in preferred_aliases:
            # it's safe if the character might be confusable with homoglyphs from other
            # categories than our preferred categories (=aliases)
            continue
        found = confusables_data.get(char, False)
        if found is False:
            continue
        # character λ is considered confusable if λ can be confused with a character from
        # preferred_aliases, e.g. if 'LATIN', 'ρ' is confusable with 'p' from LATIN.
        # if 'LATIN', 'Γ' is not confusable because in all the characters confusable with Γ,
        # none of them is LATIN.
        if preferred_aliases:
            potentially_confusable = []
            try:
                for d in found:
                    aliases = [alias(glyph) for glyph in d['c']]
                    for a in aliases:
                        if a in preferred_aliases:
                            potentially_confusable = found
                            raise Found()
            except Found:
                pass
        else:
            potentially_confusable = found
        if potentially_confusable:  # we found homoglyphs
            output = {
                'character': char,
                'alias': char_alias,
                'homoglyphs': potentially_confusable,
            }
            if not greedy:
                return [output]
            outputs.append(output)

    return outputs or False


def is_dangerous(string, preferred_aliases=[]):
    """Checks if ``string`` can be dangerous, i.e. is it not only mixed-scripts
    but also contains characters from other scripts than the ones in ``preferred_aliases``
    that might be confusable with characters from scripts in ``preferred_aliases``

    For ``preferred_aliases`` examples, see ``is_confusable`` docstring.

    >>> bool(confusables.is_dangerous('Allo'))
    False
    >>> bool(confusables.is_dangerous('AlloΓ', preferred_aliases=['latin']))
    False
    >>> bool(confusables.is_dangerous('Alloρ'))
    True
    >>> bool(confusables.is_dangerous('AlaskaJazz'))
    False
    >>> bool(confusables.is_dangerous('ΑlaskaJazz'))
    True

    :param string: A unicode string
    :type string: str
    :param preferred_aliases: Script blocks aliases which we don't want ``string``'s characters
        to be confused with.
    :type preferred_aliases: list(str)
    :return: Is it dangerous.
    :rtype: bool
    """
    return is_mixed_script(string) and is_confusable(string, preferred_aliases=preferred_aliases)



# -*- coding: utf-8 -*-
from .utils import load


categories_data = load('categories.json')


def aliases_categories(chr):
    """Retrieves the script block alias and unicode category for a unicode character.

    >>> categories.aliases_categories('A')
    ('LATIN', 'L')
    >>> categories.aliases_categories('τ')
    ('GREEK', 'L')
    >>> categories.aliases_categories('-')
    ('COMMON', 'Pd')

    :param chr: A unicode character
    :type chr: str
    :return: The script block alias and unicode category for a unicode character.
    :rtype: (str, str)
    """
    l = 0
    r = len(categories_data['code_points_ranges']) - 1
    c = ord(chr)

    # binary search
    while r >= l:
        m = (l + r) // 2
        if c < categories_data['code_points_ranges'][m][0]:
            r = m - 1
        elif c > categories_data['code_points_ranges'][m][1]:
            l = m + 1
        else:
            return (
                categories_data['iso_15924_aliases'][categories_data['code_points_ranges'][m][2]],
                categories_data['categories'][categories_data['code_points_ranges'][m][3]])
    return 'Unknown', 'Zzzz'


def alias(chr):
    """Retrieves the script block alias for a unicode character.

    >>> categories.alias('A')
    'LATIN'
    >>> categories.alias('τ')
    'GREEK'
    >>> categories.alias('-')
    'COMMON'

    :param chr: A unicode character
    :type chr: str
    :return: The script block alias.
    :rtype: str
    """
    a, _ = aliases_categories(chr)
    return a


def category(chr):
    """Retrieves the unicode category for a unicode character.

    >>> categories.category('A')
    'L'
    >>> categories.category('τ')
    'L'
    >>> categories.category('-')
    'Pd'

    :param chr: A unicode character
    :type chr: str
    :return: The unicode category for a unicode character.
    :rtype: str
    """
    _, a = aliases_categories(chr)
    return a


def unique_aliases(string):
    """Retrieves all unique script block aliases used in a unicode string.

    >>> categories.unique_aliases('ABC')
    {'LATIN'}
    >>> categories.unique_aliases('ρAτ-')
    {'GREEK', 'LATIN', 'COMMON'}

    :param string: A unicode character
    :type string: str
    :return: A set of the script block aliases used in a unicode string.
    :rtype: (str, str)
    """
    cats = [alias(c) for c in string]
    return set(cats)



# -*- coding: utf-8 -*-
import json
import os
import sys

PACKAGE_DIR = os.path.abspath(os.path.dirname(__file__))

if sys.version_info < (3,):
    import codecs
    from urllib2 import urlopen

    def u(x):
        return codecs.unicode_escape_decode(x)[0]

    def get(url, timeout=None):
        return urlopen(url, timeout=timeout)

else:
    from urllib.request import urlopen

    def u(x):
        return x

    def get(url, timeout=None):
        return urlopen(url, timeout=timeout).read().decode('utf-8').split('\n')


def path(filename):
    """Returns a file path relative to this package directory.

    :return: A file path string.
    :rtype: str
    """
    return os.path.join(PACKAGE_DIR, filename)


def load(filename):
    """Loads a JSON data file.

    :return: A dict.
    :rtype: dict
    """
    with open(path(filename), 'r') as file:
        return json.load(file)


def dump(filename, data):
    with open(path(filename), 'w+') as file:
        return json.dump(data, file)


def delete(filename):
    """Deletes a JSON data file if it exists.
    """
    try:
        os.remove(path(filename))
    except OSError:
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest

from confusable_homoglyphs import categories

latin_a = u'A'
greek_a = u'Α'

is_good = 'Allo'
looks_good = is_good.replace(latin_a, greek_a)


class TestCategories(unittest.TestCase):

    def test_aliases_categories(self):
        self.assertEqual(categories.aliases_categories(latin_a), (
            categories.alias(latin_a), categories.category(latin_a)))
        self.assertEqual(categories.aliases_categories(greek_a), (
            categories.alias(greek_a), categories.category(greek_a)))

    def test_alias(self):
        self.assertEqual(categories.alias(latin_a), 'LATIN')
        self.assertEqual(categories.alias(greek_a), 'GREEK')

    def test_category(self):
        self.assertEqual(categories.category(latin_a), 'L')
        self.assertEqual(categories.category(greek_a), 'L')

    def test_unique_aliases(self):
        self.assertEqual(categories.unique_aliases(is_good), set(['LATIN']))
        self.assertEqual(categories.unique_aliases(looks_good), set(['GREEK', 'LATIN']))


if __name__ == '__main__':
    unittest.main()



from os.path import isfile
import unittest

try:
    from urllib.error import URLError
except ImportError:
    from urllib2 import URLError

from confusable_homoglyphs.cli import generate_categories, generate_confusables
from confusable_homoglyphs.utils import get, delete, path


def unicode_org_down():
    try:
        get('ftp://ftp.unicode.org/Public', timeout=5)
        return False
    except URLError:
        return True


@unittest.skipIf(unicode_org_down(), 'www.unicode.org is down')
class TestUpdate(unittest.TestCase):
    def test_generate_categories(self):
        delete('categories.json')
        self.assertFalse(isfile(path('categories.json')))

        generate_categories()
        self.assertTrue(isfile(path('categories.json')))

    def test_generate_confusables(self):
        delete('confusables.json')
        self.assertFalse(isfile(path('confusables.json')))

        generate_confusables()
        self.assertTrue(isfile(path('confusables.json')))



# -*- coding: utf-8 -*-
import re
from collections import defaultdict

from .utils import u, get, dump

try:
    import click
except ImportError:
    print('Install this package with the [cli] extras to enable the CLI.')
    raise


@click.group()
def cli():
    pass


@cli.command()
def update():
    """
    Update the homoglyph data files from https://www.unicode.org
    """
    generate_categories()
    generate_confusables()


def generate_categories():
    """Generates the categories JSON data file from the unicode specification.

    :return: True for success, raises otherwise.
    :rtype: bool
    """
    # inspired by https://gist.github.com/anonymous/2204527
    code_points_ranges = []
    iso_15924_aliases = []
    categories = []

    match = re.compile(r'([0-9A-F]+)(?:\.\.([0-9A-F]+))?\W+(\w+)\s*#\s*(\w+)',
                       re.UNICODE)

    url = 'ftp://ftp.unicode.org/Public/UNIDATA/Scripts.txt'
    file = get(url)
    for line in file:
        p = re.findall(match, line)
        if p:
            code_point_range_from, code_point_range_to, alias, category = p[0]
            alias = u(alias.upper())
            category = u(category)
            if alias not in iso_15924_aliases:
                iso_15924_aliases.append(alias)
            if category not in categories:
                categories.append(category)
            code_points_ranges.append((
                int(code_point_range_from, 16),
                int(code_point_range_to or code_point_range_from, 16),
                iso_15924_aliases.index(alias), categories.index(category))
            )
    code_points_ranges.sort()

    categories_data = {
        'iso_15924_aliases': iso_15924_aliases,
        'categories': categories,
        'code_points_ranges': code_points_ranges,
    }

    dump('categories.json', categories_data)


def generate_confusables():
    """Generates the confusables JSON data file from the unicode specification.

    :return: True for success, raises otherwise.
    :rtype: bool
    """
    url = 'ftp://ftp.unicode.org/Public/security/latest/confusables.txt'
    file = get(url)
    confusables_matrix = defaultdict(list)
    match = re.compile(r'[0-9A-F ]+\s+;\s*[0-9A-F ]+\s+;\s*\w+\s*#'
                       r'\*?\s*\( (.+) → (.+) \) (.+) → (.+)\t#',
                       re.UNICODE)
    for line in file:
        p = re.findall(match, line)
        if p:
            char1, char2, name1, name2 = p[0]
            confusables_matrix[char1].append({
                'c': char2,
                'n': name2,
            })
            confusables_matrix[char2].append({
                'c': char1,
                'n': name1,
            })

    dump('confusables.json', dict(confusables_matrix))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import versioneer
from codecs import open
from os import path
from setuptools import setup, find_packages


def test_suite():
    test_loader = unittest.TestLoader()
    suite = test_loader.discover('tests', pattern='test_*.py')
    return suite


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open(path.join(here, 'HISTORY.rst'), encoding='utf-8') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
]

test_requirements = [
]

setup(
    name='confusable_homoglyphs',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Detect confusable usage of unicode homoglyphs, prevent homograph attacks.',
    long_description=readme + '\n\n' + history,
    author='Victor Felder',
    author_email='victorfelder@gmail.com',
    url='https://github.com/vhf/confusable_homoglyphs',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    package_data={'confusable_homoglyphs': ['categories.json', 'confusables.json']},
    entry_points={
        'console_scripts': [
            'confusable_homoglyphs=confusable_homoglyphs.cli:cli',
        ],
    },
    extras_require={
        'cli': ['click'],
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    keywords=['confusable', 'homoglyph', 'attack', 'homograph', 'unicode', 'spoofing'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Text Processing :: Filters',
        'Topic :: Text Processing',
        'Topic :: Utilities',
    ],
    test_suite='setup.test_suite',
    tests_require=test_requirements
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest

from confusable_homoglyphs import confusables

latin_a = u'A'
greek_a = u'Α'

is_good = 'Allo'
looks_good = is_good.replace(latin_a, greek_a)


class TestConfusables(unittest.TestCase):

    def test_is_mixed_script(self):
        self.assertTrue(confusables.is_mixed_script(looks_good))
        self.assertTrue(confusables.is_mixed_script(u' ρττ a'))

        self.assertFalse(confusables.is_mixed_script(is_good))
        self.assertFalse(confusables.is_mixed_script(u'ρτ.τ'))
        self.assertFalse(confusables.is_mixed_script(u'ρτ.τ '))

        try:
            confusables.is_mixed_script('', allowed_aliases=[u'COMMON'])
        except TypeError:
            self.fail('TypeError when allowed_aliases provided as unicode')

    def test_is_confusable(self):
        greek = confusables.is_confusable(looks_good, preferred_aliases=['latin'])
        self.assertEqual(greek[0]['character'], greek_a)
        self.assertIn({'c': 'A', 'n': 'LATIN CAPITAL LETTER A'}, greek[0]['homoglyphs'])
        latin = confusables.is_confusable(is_good, preferred_aliases=['latin'])
        self.assertFalse(latin)

        self.assertFalse(confusables.is_confusable(u'AlloΓ', preferred_aliases=['latin']))

        # stop at first confusable character
        self.assertEqual(len(confusables.is_confusable(u'Αlloρ', greedy=False)), 1)
        # find all confusable characters
        # Α (greek), l, o, and ρ can be confused with other unicode characters
        self.assertEqual(len(confusables.is_confusable(u'Αlloρ', greedy=True)), 4)
        # Only Α (greek) and ρ (greek) can be confused with a latin character
        self.assertEqual(
            len(confusables.is_confusable(u'Αlloρ', greedy=True, preferred_aliases=['latin'])), 2)

        # for 'Latin' readers, ρ is confusable!    ↓
        confusable = confusables.is_confusable(u'paρa', preferred_aliases=['latin'])[0]['character']
        self.assertEqual(confusable, u'ρ')
        # for 'Greek' readers, p is confusable!  ↓
        confusable = confusables.is_confusable(u'paρa', preferred_aliases=['greek'])[0]['character']
        self.assertEqual(confusable, 'p')

        try:
            confusables.is_confusable('', preferred_aliases=[u'latin'])
        except TypeError:
            self.fail('TypeError when preferred_aliases provided as unicode')

    def test_dangerous(self):
        self.assertTrue(confusables.is_dangerous(looks_good))
        self.assertTrue(confusables.is_dangerous(u' ρττ a'))
        self.assertTrue(confusables.is_dangerous(u'ρττ a'))
        self.assertTrue(confusables.is_dangerous(u'Alloτ'))
        self.assertTrue(confusables.is_dangerous(u'www.micros﻿оft.com'))
        self.assertTrue(confusables.is_dangerous(u'www.Αpple.com'))
        self.assertTrue(confusables.is_dangerous(u'www.faϲebook.com'))
        self.assertFalse(confusables.is_dangerous(u'AlloΓ', preferred_aliases=['latin']))
        self.assertFalse(confusables.is_dangerous(is_good))
        self.assertFalse(confusables.is_dangerous(u' ρτ.τ'))
        self.assertFalse(confusables.is_dangerous(u'ρτ.τ'))

    def test_char_not_found(self):
        self.assertFalse(confusables.is_dangerous(u'Їt', preferred_aliases=['latin']))
        self.assertTrue(confusables.is_dangerous(u'Їtτ', preferred_aliases=['latin']))


if __name__ == '__main__':
    unittest.main()
